/* 
 * Copyright (c) 2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * Copyright (c) 1999 Apple Computer, Inc.  All rights reserved. 
 *
 * HISTORY
 *
 */

//#define CFRUNLOOP_NEW_API 1

//#include <stdio.h>
//#include <bsd/string.h>
//#include <CoreFoundation/CoreFoundation.h>
//#include <CoreFoundation/CFMachPort.h>

//#include <libkern/OSByteOrder.h>
//#include <IOKit/IOKitLib.h>
//#include "IOUSBLib.h"
//#include <IOKit/usb/USB.h>
//#include <IOKit/usb/IOUSBUserClient.h>

//#include <mach/mach_interface.h>
//#include <IOKit/iokitmig.h> 	// mig generated
//#include <IOKit/IOKitLibPrivate.h>


__private_extern__
const char _IOKIT_USB_IOUSBLIB_C_subproj_workaround[] = "_IOKIT_USB_IOUSBLIB_C_";
